#include <stdio.h>
#include <math.h>
#include <stdlib.h>

#include "cuda_utils.h"

// CUDA Kernel //

const int blocks=90;

void __global__ ScalarProduct(int* a,int* b,int* result,int size) {
  __shared__ int blockSum;
  if (threadIdx.x==0) {
    blockSum=0;
  }
  __syncthreads();
  int threadSum=0;
  for(int i=threadIdx.x+blockIdx.x*blockDim.x; i<size; 
      i+=blockDim.x*gridDim.x) {
    threadSum+=a[i]*b[i];
  }
  // TODO: add both reduction stages here //
}

// CPU code //

int main() {

  srand48(time(NULL));

  // define vector size //
  const int vector_size = 10000;

  // allocate memory on cpu //
  int* vector_a_cpu;
  int* vector_b_cpu;
  cu_verify(cudaMallocHost((void**)&vector_a_cpu,vector_size*sizeof(int)));
  cu_verify(cudaMallocHost((void**)&vector_b_cpu,vector_size*sizeof(int)));
  int result_cpu = 0;

  // initialize vectors //
  for(int i=0;i<vector_size;i++) {
    vector_a_cpu[i]=int(drand48()*100);
    vector_b_cpu[i]=int(drand48()*100);
  }

  // allocate memory on gpu //
  int* vector_a_gpu;
  int* vector_b_gpu;
  int* result_gpu;
  cu_verify(cudaMalloc((void**)&vector_a_gpu,vector_size*sizeof(int)));
  cu_verify(cudaMalloc((void**)&vector_b_gpu,vector_size*sizeof(int)));
  cu_verify(cudaMalloc((void**)&result_gpu,sizeof(int)));

  // copy input vector from cpu to gpu //
  cu_verify(cudaMemcpy(vector_a_gpu,vector_a_cpu,vector_size*sizeof(int),
                       cudaMemcpyHostToDevice));
  cu_verify(cudaMemcpy(vector_b_gpu,vector_b_cpu,vector_size*sizeof(int),
                       cudaMemcpyHostToDevice));
  cu_verify(cudaMemset(result_gpu,0,sizeof(int)));

  Timer timer;
  initTimer(&timer);

  // cuda kernel call //
  cu_verify_void((ScalarProduct<<<blocks,32>>>(vector_a_gpu,vector_b_gpu,
                                               result_gpu,vector_size)));
  cu_verify(cudaThreadSynchronize());
  double duration = getTimer(&timer);

  // copy result from gpu to cpu //
  cu_verify(cudaMemcpy(&result_cpu,result_gpu,sizeof(int),
                       cudaMemcpyDeviceToHost));

  // verify result //
  int sum=0;
  for(int i=0;i<vector_size;i++) {
    sum+=vector_a_cpu[i]*vector_b_cpu[i];
  }
  if (fabsf(sum-result_cpu)>1e-5) {
    printf("computation result is wrong: expected=%i cuda result=%i\n",
           sum,result_cpu);
  } else {
    printf("computation result is correct.\n");
  }

  printf("duration=%le seconds, performance = %le FLOPS\n",
         duration,2.*vector_size/duration);

  // free gpu memory //
  cu_verify(cudaFree(vector_a_gpu));
  cu_verify(cudaFree(vector_b_gpu));
  cu_verify(cudaFree(result_gpu));

  // free cpu memory //
  cu_verify(cudaFreeHost(vector_a_cpu));
  cu_verify(cudaFreeHost(vector_b_cpu));

  // exit //
  return 0;
}
