module device_code

! block size 'T' (width and height of the shared memory block) !
integer, parameter :: blockSize=16

contains

! CUDA Kernel !

attributes(global) subroutine MatrixProd(a,b,c,m,n,p)
  implicit none
  integer, value :: m,n,p
  integer :: index_i,index_j,index_k_coarse,index_k_fine
  integer :: column_a,row_b,end
  real, dimension(0:m*n-1), intent(in) :: a
  real, dimension(0:n*p-1), intent(in) :: b
  real, dimension(0:m*p-1), intent(inout) :: c
  real, shared, dimension(0:blockSize*(blockSize+1)-1) :: shared_a,shared_b
  real :: sum

  index_i=(blockIdx%y-1)*blockSize+threadIdx%y-1
  index_j=(blockIdx%x-1)*blockSize+threadIdx%x-1

  sum=0
  ! iterate over blocks !
  do index_k_coarse=0,n-1,blockSize
    ! read block of a into shared memory !
    column_a=threadIdx%x-1+index_k_coarse
    if (column_a<n .and. index_i<m) then
      shared_a(threadIdx%x-1+blockSize*(threadIdx%y-1))=a(index_i*n+column_a)
    end if
    ! read block of b into shared memory !
    row_b=threadIdx%y-1+index_k_coarse
    if (index_j<p .and. row_b<n) then
      shared_b(threadIdx%x-1+blockSize*(threadIdx%y-1))=b(row_b*p+index_j)
    end if
    call syncthreads()
    ! compute sum for one block !
    if (index_j<p .and. index_i<m) then
      end=min(blockSize,n-index_k_coarse)-1
      do index_k_fine=0,end
        sum=sum+shared_a(index_k_fine+blockSize*(threadIdx%y-1)) &
            *shared_b(threadIdx%x-1+blockSize*(index_k_fine))
      end do
    end if
    call syncthreads()
  end do
  ! assign result to matrix c !
  if (index_j<p .and. index_i<m) then
    c(index_i*p+index_j)=sum
  end if
end subroutine MatrixProd

end module device_code

module cpu_code

contains

! CPU code fore reference computation !

subroutine MatrixProdReference(a,b,c,m,n,p)
  integer, value :: m,n,p
  integer :: index_i,index_j,index_k,a_line_index
  real, dimension(0:m*n-1), intent(in) :: a
  real, dimension(0:n*p-1), intent(in) :: b
  real, dimension(0:m*p-1), intent(out) :: c
  real :: sum
  do index_i=0,m-1
    a_line_index=index_i*n
    do index_j=0,p-1
      sum=0
      do index_k=0,n-1
        sum=sum+a(a_line_index+index_k)*b(index_k*p+index_j)
      end do
      c(index_i*p+index_j)=sum
    end do
  end do
end subroutine MatrixProdReference

end module cpu_code

program exercise08

use cudafor
use device_code
use cpu_code

implicit none

! define vector size !
integer, parameter :: size_m = 450
integer, parameter :: size_n = 350
integer, parameter :: size_p = 400
integer, parameter :: matrix_a_size = size_m*size_n
integer, parameter :: matrix_b_size = size_n*size_p
integer, parameter :: matrix_c_size = size_m*size_p
!real, dimension(0:matrix_a_size-1) :: matrix_a_cpu
!real, dimension(0:matrix_b_size-1) :: matrix_b_cpu
!real, dimension(0:matrix_c_size-1) :: matrix_c_cpu
real, pinned, allocatable :: matrix_a_cpu(:)
real, pinned, allocatable :: matrix_b_cpu(:)
real, pinned, allocatable :: matrix_c_cpu(:)
real, dimension(0:matrix_c_size-1) :: matrix_c_reference
real, dimension(0:matrix_a_size-1), device :: matrix_a_gpu
real, dimension(0:matrix_b_size-1), device :: matrix_b_gpu
real, dimension(0:matrix_c_size-1), device :: matrix_c_gpu
type(dim3) :: block_dim,grid_dim
real(kind=8) :: timer_transfer_in,timer_kernel,timer_transfer_out
real(kind=8) :: timer_end,timer_begin_cpu,timer_end_cpu,duration
integer :: i,j
logical :: success
integer :: error
character(len=256):: string

! CPU code !

call random_seed()

allocate(matrix_a_cpu(0:matrix_a_size-1))
allocate(matrix_b_cpu(0:matrix_b_size-1))
allocate(matrix_c_cpu(0:matrix_c_size-1))

! initialize cpu input matrices to random values !
do i=0,matrix_a_size-1
  call random_number(matrix_a_cpu(i))
end do
do i=0,matrix_b_size-1
  call random_number(matrix_b_cpu(i))
end do
! initialize cpu output matrix to -1 !
matrix_c_cpu=-1

! initialize gpu output vector !
matrix_c_gpu=matrix_c_cpu

call cpu_time(timer_transfer_in)

! copy input vectors a,b to gpu !
matrix_a_gpu=matrix_a_cpu
matrix_b_gpu=matrix_b_cpu

call cpu_time(timer_kernel)

block_dim=dim3(blockSize,blockSize,1)
grid_dim=dim3((size_p+blockSize-1)/blockSize,(size_m+blockSize-1)/blockSize,1)

! cuda kernel call !
call MatrixProd<<<grid_dim,block_dim>>>(matrix_a_gpu,matrix_b_gpu,matrix_c_gpu,size_m,size_n,size_p)
error = cudaThreadSynchronize()
call cpu_time(timer_transfer_out)
string=cudaGetErrorString(error)
print *,trim(string)

! copy output vector from gpu to cpu !
matrix_c_cpu=matrix_c_gpu
call cpu_time(timer_end)

call cpu_time(timer_begin_cpu)
call MatrixProdReference(matrix_a_cpu,matrix_b_cpu,matrix_c_reference,size_m,size_n,size_p);
call cpu_time(timer_end_cpu)

! verify results !
success=.true.
LOOP_I: do i=0,size_m-1
  do j=0,size_p-1
    if (abs(matrix_c_cpu(i*size_p+j)-matrix_c_reference(i*size_p+j)) &
      > 1e-5 * abs(matrix_c_cpu(i*size_p+j)+matrix_c_reference(i*size_p+j))) then
      print *,'error: position=',i,',',j,' reference=',matrix_c_reference(i*size_p+j),' cuda result=',matrix_c_cpu(i*size_p+j)
      success=.false.
      exit LOOP_I
    end if
  end do
end do LOOP_I

if (success) then
  print *,"computation result is correct."
end if

duration=timer_kernel-timer_transfer_in
print *,'duration(data transfern in)=',duration,' seconds,  bandwidth = ',4.*(matrix_a_size+matrix_b_size)/duration,' bytes/s'
duration=timer_end-timer_transfer_out
print *,'duration(data transfern out)=',duration,' seconds,  bandwidth = ',4.*(matrix_c_size)/duration,' bytes/s'
duration=timer_transfer_out-timer_kernel
print *,'duration(computation)=',duration,' seconds, performance = ',2.*size_m*size_n*size_p/duration,' FLOPS'
duration=timer_end-timer_transfer_in
print *,'duration(overall)=',duration,' seconds, performance(including transfer time) = ',2.*size_m*size_n*size_p/duration,' FLOPS'
duration=timer_end_cpu-timer_begin_cpu
print *,'duration(computation on CPU)=',duration,' seconds, performance = ',2.*size_m*size_n*size_p/duration,' FLOPS'

end program exercise08
