#include <stdio.h>
#include <math.h>

#include "cuda_utils.h"

// CUDA Kernel //

const int blocks=90;

void __global__ ScalarProduct1(float* a,float* b,float* result,int size) {
  __shared__ float sums[16];
  float sum=0;
  for(int i=threadIdx.x+blockIdx.x*blockDim.x; i<size; 
      i+=blockDim.x*gridDim.x) {
    sum+=a[i]*b[i];
  }
  __syncthreads();
  if (threadIdx.x<16)
    sums[threadIdx.x]=sum;
  __syncthreads();
  if (threadIdx.x>=16)
    sums[threadIdx.x-16]+=sum;
  __syncthreads();
  if (threadIdx.x<8)
    sums[threadIdx.x]+=sums[threadIdx.x+8];
  __syncthreads();
  if (threadIdx.x<4)
    sums[threadIdx.x]+=sums[threadIdx.x+4];
  __syncthreads();
  if (threadIdx.x==0)
    result[blockIdx.x]=(sums[0]+sums[1])+(sums[2]+sums[3]);
}

void __global__ ScalarProduct2(float* input,float* output) {
  float sum=0;
  for(int i=0;i<blocks;i++) {
    sum+=input[i];
  }
  *output=sum;
}

// CPU code //

int main() {

  cu_verify(cudaSetDeviceFlags(cudaDeviceMapHost));

  // define vector size //
  const int vector_size = 10000;

  // allocate memory on cpu //
  float* vector_a_cpu;
  float* vector_b_cpu;
  cu_verify(cudaMallocHost((void**)&vector_a_cpu,vector_size*sizeof(float)));
  cu_verify(cudaMallocHost((void**)&vector_b_cpu,vector_size*sizeof(float)));
  float* result_cpu;
  cu_verify(cudaHostAlloc((void**)&result_cpu,sizeof(float),
                          cudaHostAllocMapped));
  *result_cpu=0;

  // initialize vectors //
  for(int i=0;i<vector_size;i++) {
    vector_a_cpu[i]=1.f/(1.f+i);
    vector_b_cpu[i]=1.f+i;
  }

  // allocate memory on gpu //
  float* vector_a_gpu;
  float* vector_b_gpu;
  float* intermediate_results_gpu;
  float* result_gpu;
  cu_verify(cudaMalloc((void**)&vector_a_gpu,vector_size*sizeof(float)));
  cu_verify(cudaMalloc((void**)&vector_b_gpu,vector_size*sizeof(float)));
  cu_verify(cudaMalloc((void**)&intermediate_results_gpu,
                       blocks*sizeof(float)));
  cu_verify(cudaHostGetDevicePointer((void**)&result_gpu,result_cpu,0));

  // copy input vector from cpu to gpu //
  cu_verify(cudaMemcpy(vector_a_gpu,vector_a_cpu,vector_size*sizeof(float),
                       cudaMemcpyHostToDevice));
  cu_verify(cudaMemcpy(vector_b_gpu,vector_b_cpu,vector_size*sizeof(float),
                       cudaMemcpyHostToDevice));

  Timer timer;
  initTimer(&timer);

  // cuda kernel call //
  cu_verify_void((ScalarProduct1<<<blocks,32>>>(vector_a_gpu,vector_b_gpu,
                                                intermediate_results_gpu,
                                                vector_size)));
  cu_verify_void((ScalarProduct2<<<1,1>>>(intermediate_results_gpu,
                                          result_gpu)));

  cu_verify(cudaThreadSynchronize());
  double duration = getTimer(&timer);

  // verify result //
  float sum=0;
  for(int i=0;i<vector_size;i++) {
    sum+=vector_a_cpu[i]*vector_b_cpu[i];
  }
  if (fabsf(sum-*result_cpu)>1e-5) {
    printf("computation result is wrong: expected=%g cuda result=%g\n",
           sum,*result_cpu);
  } else {
    printf("computation result is correct.\n");
  }

  printf("duration=%le seconds, performance = %le FLOPS\n",
         duration,2.*vector_size/duration);

  // free gpu memory //
  cu_verify(cudaFree(vector_a_gpu));
  cu_verify(cudaFree(vector_b_gpu));
  cu_verify(cudaFree(result_gpu));

  // free cpu memory //
  cu_verify(cudaFreeHost(vector_a_cpu));
  cu_verify(cudaFreeHost(vector_b_cpu));

  // exit //
  return 0;
}
